{$A+ Align data on word boundary }
{$F- Don't force FAR calls       }
{$B- Short circuit boolean eval  }
{$V- Relax string checking       }

{$I- Enable System.IOResult error checks }
{$N+,E+ Allow for software 80x87 }
{$R- Range checking off }
{$S- Stack checking off }

{$L+ Enable gen. of local symbols }
{$D+ Enable debug information     }
{$O- No overlays                  }

{$M  4096,0, 655360} { Max-Stack Low-heap High-heap }

PROGRAM LNGapply; { (C) Copyright Adolfo Di Mare, 1994 }

{ NOTICE
  Use this programs at your own risk. The author
  reserves all comercial rights. This program, its
  source code and documentation have been put in the
  public domain.                                     }

{ USAGE:
  LNGapply [lngfile[.ext]] [LNG(language)]

  Takes the translated line groups from file
  [lngfile.[ext]] and applies them to the source
  file.

  - lngfile is obtained running program LNGrep, and
    editing the resulting file.

  - LNGapply contains a moderate amount of error
    checking to protect the programmer from
    him/herself. However, when the line numbers
    for each group get messed up, the original
    source files can be damaged to no recognition.

  - When lngfile.ext is not included in the command
    line, file name TEXTFILE is used as default.

  - The original source files are retained with a
    .bak extention.                                 }

{ EXAMPLE:
   C:> LNGapply TEXTFILE LNG(spanish)

   In this case, all the annotations in TEXTFILE are
   applied to the source files. The annotation that
   leads each group of annotated line will be the
   Spanish annotation.                              }

{ PROGRAMMER                      Adolfo Di Mare    }

USES
  Dos, Crt, LngTools;

PROCEDURE WriteUSAGE;
{ RESULT
  Displays a help message. }
BEGIN { WriteUSAGE }
  WriteLn('LNGapply v', LngTools.LNG_version:1);
  WriteLn('USAGE: ');
  WriteLn('LNGapply [lngfile[.ext]] [LNG(language)]');
  WriteLn;
  WriteLn(
  'Takes the translated line groups from file');
  WriteLn(
  '[lngfile.[ext]] and applies them to the source');
  WriteLn(
  'file.');
END; { WriteUSAGE }


PROCEDURE Copy_File(              { ADOLFO }
  {?} VAR inF  : FILE_OF_CHAR_T;
  {?} VAR outF : TEXT
);
{ RESULT
  Copies the contents of "inF" into "outF".
  - Closes both files upon termination.    }
VAR
  str : STRING;
BEGIN { Copy_File }
  WHILE NOT Eof(inF) DO BEGIN
    IF Get_Line(inF, str) THEN BEGIN
      WriteLn(outF, str);
    END
    ELSE BEGIN
      WriteLn(outF, ERROR_line);
    END;
  END;

  { finish up }
  Close(inF);
  Close(outF);
END; { Copy_File }

PROCEDURE Advance_to_String_Group(   { ADOLFO }
  {?} VAR SG   : TString_Group;
  {+} VAR lng  : LNG_T;
  {?} VAR inF  : FILE_OF_CHAR_T;
  {?} VAR outF : TEXT;
  {?} VAR line : WORD;
  {+}     from : WORD;
  {+}     count: WORD
);
{ RESULT
  Keeps copying every line from "inF" into
  "outF" until line number "from" is reached.
  Then, the string group "SG" gets written
  out replacing the "count" lines that form a
  string group in "inF".
  - It is presummed that exactly "line" lines
    have been read in "inF". line=0 is ok.
  - "SG" becomes empty.
  - It is posible for this procedure not to
    write any lines at all.
  - "line" is updated to reflect the fact
    that the group was copied into "outF".    }
VAR
 i  : WORD;
 str: STRING;
BEGIN { Advance_to_String_Group }
  { avoid the MaxInt loop }
  IF SG.Empty OR (from = 0) THEN BEGIN
    EXIT;
  END;

  { advance to the beggining of the group }
  FOR i := line+1 TO from-1 DO BEGIN
    IF Get_Line(inF, str) THEN BEGIN
      WriteLn(outF, str);
    END
    ELSE BEGIN
      WriteLn(outF, ERROR_line);
    END;
  END;

  { skip the group lines within "prg" }
  FOR i := 1 TO count DO BEGIN
    IF NOT Get_Line(inF, str) THEN BEGIN
      { ignore these errors }
    END;
  END;

  { synchronize "line" with "count" in "inF" }
  line := from + count - 1;

  SG.Write(outF, lng);
END; { Advance_to_String_Group }

VAR
  SG : TString_Group;
  { SG is so big that it would blow up the run }
  { time stack if it were a local variable...  }

PROCEDURE Apply_File(                  { ADOLFO }
  {+} def_lng  : LNG_T;    { default annotation }
  {+} lng_fname: STRING    { file to process    }
);
{ RESULT
  The files referenced in "lng_fname" are
  processed line by line, and those lines that
  correspond to an annotated line get substited
  by their corresponding line in "lng_fname".

  - "lng_fname" is a text file produced by
    program LNGrep. It contains groups of
    annotated lines, each of which corresponds
    to a different language.

  - File "lng_fname" contains a reference to
    each of the annotated source files that
    correspond to each of its lines.

  - No change is made to the original file when
    many lines in a group have the LNG() empty
    annotation, because this clearly represents
    a mistake.                                  }

{ EXAMPLE
  *=====> File [WO.PRG]        && LNG(us)
  []-------------------------------------[11:1]
  xcode = "SO"                 && LNG()
  []-------------------------------------[12:1]
  scompany = " " + s_comp      && LNG()
  []-------------------------------------[32:1]
   MMENU("Sistema de Crdito") && LNG(sp)
  *MMENU("Credit System")      && LNG(us)
  []--------------------------------------[0:0]
  *=====> File [WO1.PRG]       && LNG(us)
  []--------------------------------------[0:0] }

VAR

  lngF : FILE_OF_CHAR_T; { file named "lng_fname" }
           { input file with language annotations }

  prgF : FILE_OF_CHAR_T; { old file: this is what }
  prgF_open : BOOLEAN;   { gets copied into newF  }

  prg_fname : STRING;    { name for prgF (string) }

  newF : TEXT;  { new file, including annotations }

  line : Tokenized_T;

  skip_this_group : BOOLEAN;
  {
    TRUE when an error is detected in a group of
    lines. The lines are ignored.
  }

VAR
  i,         { looper }
  n,
  m : WORD;  { quantity holders }

  add_LNG : BOOLEAN;

VAR
  group : RECORD
    line  : WORD;  { starting line for SG }
    count : WORD;  { # of lines in group  }
  END;

  line_prg : WORD;    { current line in prgF }

VAR
  path: Dos.PathStr;  { for Dos.Fsplit() }
  dir : Dos.DirStr;
  name: Dos.NameStr;
  ext : Dos.ExtStr;
  str : STRING;
BEGIN { Apply_File }

  { open the language annotations input file  }
  Assign(lngF, lng_fname);
  Reset(lngF);

  skip_this_group := TRUE;

  prgF_open   := FALSE;
  line_prg    := 0;
  group.line  := 0;
  group.count := 0;

  SG.Init;

  { This loop gets each line from lngF }
  WHILE Get_Line(lngF, line.str) DO BEGIN

    IF IS_File_Name(line.str, prg_fname) THEN BEGIN

      skip_this_group := FALSE;

      IF prgF_open THEN BEGIN
        { copy the previous group }
        Advance_to_String_Group(
          SG, def_lng, prgF, newF,
          line_prg, group.line, group.count);

        { copy remainder of prgF into newF }
        Copy_File(prgF, newF);
      END;

      { parse the file name }
      Dos.Fsplit(prg_fname, path, name, ext);

      { create the .bak file }
      path := path+name;
      Rename_to_BAK(path, ext);

      { open the two new files }
      Assign(prgF, path + '.bak');
      Reset(prgF);
      { the input file is source.bak file  }

      { the new file where the annotations }
      { will be included is created        }
      Assign(newF, prg_fname);
      ReWrite(newF);  { old name, new file }

      prgF_open   := TRUE;
      line_prg    := 0;
      group.line  := 0;
      group.count := 0;

    END

    ELSE IF skip_this_group THEN BEGIN
      { keep reading until a dashed line comes in }
    END

    ELSE IF IS_Blank(line.str) THEN BEGIN
      { ignore blank lines }
    END

    ELSE BEGIN
      { add another line to the group }

      IF NOT IS_Dashed(line.str, n, m) THEN BEGIN
        { add the line to the group }
        Tokenize(
          line.str, line.prs, line.com, line.lng);
        SG.Insert(line, n);
      END
      ELSE BEGIN
        { all the lines in the group have been read }
        { proceed to include the lines into prgF    }

        IF prgF_open THEN BEGIN
          { copy the previous group }
          Advance_to_String_Group(
            SG, def_lng, prgF, newF,
            line_prg, group.line, group.count);
        END;

        { test the new group count limits }
        IF
          (n <= group.line) OR
          (m <= 0)          OR
          (m > DIM_SG)
        THEN BEGIN
          { an error ocurred because line }
          { numbers are not increasing or }
          { the line count is invalid.    }

          { go on to the next group. }
          { don't change "group"     }
          skip_this_group := TRUE;
        END
        ELSE BEGIN
          { keep count limits for next group }
          group.line  := n;
          group.count := m;
          skip_this_group := FALSE;
        END;

      END;

    END; { add another line to the group }

  END; { main loop }

  IF prgF_open THEN BEGIN
    { copy the previous group }
    Advance_to_String_Group(
      SG, def_lng, prgF, newF,
      line_prg, group.line, group.count);

    { copy the tail of the file }
    Copy_File(prgF, newF);
  END;

  Close(lngF);
  SG.Done;
END; { Apply_File }


VAR
  def_lng  : LNG_T;     { default annotation }
  lng_fname: STRING;    { file to process    }

  ignore: BOOLEAN;

BEGIN { LNGapply }

  { get the command line parameters }
  IF NOT Command_Line(
           def_lng, lng_fname, ignore, ignore)
  THEN BEGIN
    WriteUSAGE;
    EXIT;
  END;

  Apply_File(def_lng, lng_fname);

END. { LNGapply }
