{$A+ Align data on word boundary }
{$F- Don't force FAR calls       }
{$B- Short circuit boolean eval  }
{$V- Relax string checking       }

{$I- Enable System.IOResult error checks }
{$N+,E+ Allow for software 80x87 }
{$R- Range checking off }
{$S- Stack checking off }

{$L+ Enable gen. of local symbols }
{$D+ Enable debug information     }
{$O- No overlays                  }

{$M  4096,0, 655360} { Max-Stack Low-heap High-heap }

PROGRAM LNGclean; { (C) Copyright Adolfo Di Mare, 1994 }

{ NOTICE
  Use this programs at your own risk. The author
  reserves all comercial rights. This program, its
  source code and documentation have been put in the
  public domain.                                     }

{ USAGE:
  LNGclean [wildcard[.ext]] [LNG(language)] [>file]

  Scans all the source files identified by
  [wildcard.[ext]] and writes into the standard
  output all the lines in each file but those that
  have a language annotation that are different
  from LNG(language) and the language independent
  annotation LNG().

  - Every annotation is removed from all the output
    lines.

  - The output of this command goes to the standard
    output.

  - Wildcards are supported. All the translated
    are written out to the standard output. Each
    cleaned up file is preceeded by a nice asterisk
    box that has its name.

  - Standard output redirection [>file] is
    supported.                                       }


{ REMARKS
  To apply this command to a bunch of files a .bat
  file can be used in conjunction with a for loop:

    REM BatLNG.bat to LNGclean many files
    if     (%2)==() goto help
    if not (%3)==() goto help
    if not exist bak\null md bak
    LNGclean %1 %2 >E:\RamDisk\t.bak
    copy     %1 bak
    copy     E:\RamDisk\t.bak   %1
    goto end

    :help
    echo USAGE BatLNG lngfile[.ext] LNG(language)

    :end
    REM end of file BatLNG.bat

  C:> for %a in (*.prg) do call BatLNG %a lng(us)    }

{ EXAMPLE:
   C:> LNGclean short.prg LNG(sp) >short.sp

   The non annotated version of Short.prg written in
   the Spanish language version is copied into file
   "short.sp".                                       }

{ PROGRAMMER                       Adolfo Di Mare    }

USES
  Dos, Crt, LngTools;


PROCEDURE WriteUSAGE;
{ RESULT
  Displays a help message. }
BEGIN { WriteUSAGE }
  WriteLn('LNGclean v', LngTools.LNG_version:1);
  WriteLn('USAGE: ');
  Write  ('LNGclean [wildcard[.ext]]');
  WriteLn(' [LNG(language)] [>file]');
  WriteLn;
  WriteLn(
  'Scans all the source files identified by');
  WriteLn(
  '[wildcard.[ext]] and writes into the standard');
  WriteLn(
  'output all the lines in each file but those that');
  WriteLn(
  'have a language annotation that are different');
  WriteLn(
  'from LNG(language) and the language independent');
  WriteLn(
  'annotation LNG().');
END; { WriteUSAGE }

VAR
  SG : TString_Group;
  { SG is so big that it would blow up the run }
  { time stack if it were a local variable...  }


PROCEDURE Clean_File(                  { ADOLFO }
  {+} VAR outF : TEXT;            { output file }
  {+} def_lng  : LNG_T;    { default annotation }
  {+} VAR dir  : Dos.DirStr;     { Dos.Fsplit() }
  {+} VAR fname: Dos.PathStr  { file to process }
);
{ RESULT
  "fname" is a program source file, written in
  any of the xBase dialects, that contains group
  of anoted lines.

  Each group of annotated lines is identified by
  Clean_File(). The line that has an annotation
  equal to "def_lng" is the one that gets
  written out as the only non commented line.
  All the other lines in the string group are
  not written out.

  - When no annotation equal to "def_lng" is
    found in a string group, the leading line
    in the group gets written out.

  - outF must be open. It remains open upon
    return.                                     }

VAR
  prgF : FILE_OF_CHAR_T;

  str : Tokenized_T;

  ignore: STRING[LNG_size];

  i : WORD;
  line_no : WORD; { current line in "fname" }

  path: Dos.PathStr;

BEGIN { Clean_File }
{
  The logic for this procedure resembles
  very much that of program LNGswap.

  Their differences are pretty small, because
  LNGswap needs to write EVERY line in a
  string group. whereas LNGclean writes only
  the LEADING one.
}

  { write out a nice box with the file name }
  path := dir +  fname;

  Write(outF, '**');
  FOR i := 1 TO Length(path) DO BEGIN
    Write(outF, '*')
  END;
  Write(outF, '**');
  WriteLn(outF);

  Write(outF, '* ');
  Write(outF, path);
  Write(outF, ' *');
  WriteLn(outF);

  Write(outF, '**');
  FOR i := 1 TO Length(path) DO BEGIN
    Write(outF, '*')
  END;
  Write(outF, '**');
  WriteLn(outF);

  { open the two new files }
  Assign(prgF, path);
  Reset(prgF);

  { Initialize and pop empty the string group }
  SG.Init;

  line_no := 0;

  WHILE Get_Line(prgF, str.str) DO BEGIN

    INC(line_no);
    Tokenize(str.str, str.prs, str.com, str.lng);

    IF str.lng = '' THEN BEGIN
      { clear the previous group }
      SG.Write_One(outF, def_lng);

      { there is no annotation: just copy the line }
      WriteLn(outF, str.str);
    END

    ELSE IF str.lng = LNG_all THEN BEGIN
      { skip language independent strings }
      SG.Write_One(outF, def_lng);

      { copy the line }
      IF Has_Annotation(str.str, i, ignore) THEN BEGIN
        WriteLn(
          outF, RTrim(System.Copy(str.str, 1, i-1)));
      END
      ELSE BEGIN
        WriteLn(outF, str.str);
      END;
    END

    ELSE BEGIN
      { the line already has an LNG() annotation  }

      IF (NOT str.com)      OR { not a comment?  }
         SG.Empty           OR { not in a group? }
         SG.IS_Member(str.lng) { already there?  }
      THEN BEGIN

        { a new group begins: flush the old one }
        SG.Write_One(outF, def_lng);

        { now decide whether to begin a new group }
        IF NOT str.com THEN BEGIN
          SG.Insert(str, line_no);
        END
        ELSE BEGIN
          { copy the line }
          IF Has_Annotation(str.str, i, ignore)
          THEN BEGIN
            WriteLn(
              outF, RTrim(System.Copy(str.str, 1, i-1)));
          END
          ELSE BEGIN
            WriteLn(outF, str.str);
          END;
        END;

      END
      ELSE BEGIN
      {
        (str.com)      AND          it's a comment
        (NOT SG.Empty) AND          there is a group
        (NOT SG.IS_Member(str.lng)) not in the group
      }
        { add the line to the group }
        SG.Insert(str, line_no);
      END;

    END;

  END; { WHILE }

  SG.Write_One(outF, def_lng);

  IF NOT Eof(prgF) THEN BEGIN
    Write  (outF, 'LNGrep: line too long in file ');
    WriteLn(outF,  fname, ' (', line_no, ')');
    WriteLn(outF, 'LNGrep: file scanning aborted!!!');
  END;

  SG.Done;
  Close(prgF);
END; { Clean_File }


TYPE
  Turbo_SearchRec = RECORD       { from Online Help }
    fill: ARRAY[1..21] OF BYTE;
    attr: BYTE;
    time: LONGINT;
    size: LONGINT;
    name: ARRAY[0..12] OF CHAR;
  END;

VAR
  def_lng  : LNG_T;
  add_LNG  : BOOLEAN;
  ignore   : BOOLEAN;
  wildcard : STRING;

VAR
  srec   : Dos.SearchRec;
CONST
  {$IFNDEF VER55}   FName_S = SizeOf(srec.name);
  {$ELSE}           FName_S = 12;
  {$ENDIF !VER55}

TYPE
  Node_P = ^Node_T;
  Node_T = RECORD
    fname : STRING[FName_S];
    next  : Node_P;
  END;

VAR
  list,
  node,
  last   : Node_P;

VAR
  dir : Dos.DirStr;
  path: Dos.PathStr;
  name: Dos.NameStr;
  ext : Dos.ExtStr;

BEGIN { LNGclean }

  { get the command line parameters }
  IF NOT Command_Line(
           def_lng, wildcard, add_LNG, ignore)
  THEN BEGIN
    WriteUSAGE;
    EXIT;
  END;

  { redirect I/O }
  IF NOT ISaTTY(1) THEN BEGIN
    Assign(OUTPUT,'');
    ReWrite(OUTPUT);
  END;

{
  wildcard := 'test.pas';
  def_lng  := 'LNG(sp)';
}

  { parse the file name }
  Dos.Fsplit(wildcard, dir, name, ext);

  { "dir" contains the drive/directory }
  { "dir" is the path to each file     }

  { create a list with all the file names }
  Dos.FindFirst(
    wildcard,
    Dos.Archive + Dos.ReadOnly,
    srec);

  list := NIL;
  WHILE Dos.DosError = 0 DO BEGIN

    NEW(node);
    node^.fname := srec.name;

    IF list = NIL THEN BEGIN
      list := node;   { empty list }
    END
    ELSE BEGIN
      last^.next := node; { append }
    END;

    last := node;

    Dos.FindNext(srec);

  END;

  { finish up the list }
  IF list <> NIL THEN BEGIN
    last^.next := NIL;
  END;
{
  In LNGswap.pas I included an explanation of why
  I keep all the file names in a list, if you are
  interested.

  This really isn't necessary here, but I decided
  to include it just in case.
}

  { process each file }
  WHILE list <> NIL DO BEGIN

    Clean_File(OUTPUT, def_lng, dir, list^.fname);

    node := list;
    list := list^.next;
    DISPOSE(node);
  END;

END. { LNGclean }
