{$A+ Align data on word boundary }
{$F- Don't force FAR calls       }
{$B- Short circuit boolean eval  }
{$V- Relax string checking       }

{$I- Enable System.IOResult error checks }
{$N+,E+ Allow for software 80x87 }
{$R- Range checking off }
{$S- Stack checking off }

{$L+ Enable gen. of local symbols }
{$D+ Enable debug information     }
{$O- No overlays                  }

{$M  4096,0, 655360} { Max-Stack Low-heap High-heap }

PROGRAM LNGrep; { (C) Copyright Adolfo Di Mare, 1994 }

{ NOTICE
  Use this programs at your own risk. The author
  reserves all comercial rights. This program, its
  source code and documentation have been put in the
  public domain.                                     }

{ USAGE:
  LNGrep wildcard[.ext] [options] [LNG(lang)] [>file]

  Extracts from all the xBase[.prg] program files in
  the current directory all lines that contain text
  strings, and writes them to the standard output.

  After redirecting the output of this program into
  a file, names TEXTFILE in most cases, that file
  can be edited by hand, to translate each of the
  string to another language.

  All the translated string can be reinstalled in
  the original program files using program LNGapply.

  In this way, the original source programs need not
  be edited one by one. It will be more difficult for
  mistakes to creep in because LNGapply will
  systematically include the translated lines into
  the original files.

  - LNGrep handles annotated lines in groups.
    Every annotated line group begins with a not
    commented out annotated line. All the lines
    in a group are annotated lines. If the
    language annotation in the command line
    [LNG(language)] does not appear in a
    group, it is added when the group gets
    written to the standard output.

  - A group never begins with a commented out
    line.

  - LNG() is the language independent
    annotation.

  - Command line options are distinguished from
    other strings because they begin with one
    of these characters: +/-.

  - The valid command line options are "-add"
    and "-new".

  - When the "-add" option is used, a new line
    is copied to the standard output with the
    the language independent annotation [LNG()].
    This line will be a copy of the "lng"
    annotation if it exists, or a copy of the
    first line to the contrary.

    The line with the LNG() annotation is the
    one to be translated into the other
    language.

  - The "-add" option should be used only the
    first time LNGrep is run in a set of files.
    After the first time, there will seldom be
    a need for LNGrep to add LNG() annotations
    automatically.

  - The "-new" is used to instruct LNGrep to
    printout only those lines that don't belong
    in a string group, because they have no
    language annotation. This is usefull to make
    a quick translation after a few changes are
    made in the source program.

  - Lines that have the language independent
    annotation never get written to standard
    output. These lines never need translation.

  - After translation, program LNGapply is used
    to include the translated lines in the
    source files.

  - When the extracted line already contains
    an && LNG(lng) annotation, LNGrep will not
    change it.

  - "wildcard" can contains the common MSDOS
    wildcar characters "*.*" "????.*". Only one
    wildcard especification should appear in
    the command line; if more are required,
    then LNGrep can be run once for each
    wildcard specification.

  - .bak files are never processed.

  - All lines in each file must be shorter than
    LNGtools.LineLength characters.

  - LNGtools.LNG_default is the default language
    annotation.

  - Usually, TEXTFILE is the name of the file
    where the output of LNGrep is redirected.     }
{ EXAMPLE:
   C:> LNGrep -add LNG(us) *.prg

   In this case, all the Xbase program files will
   be scanned by LNGrep. Every time that a line
   that contains a string is found it is included
   in its string group, and every string group is
   written out to standard output, which is the
   console in this case.

   - Should option /new had been specified, only
     lines that have a single or double quote
     that don't belong to a string group would
     have been written out.                          }

{ PROGRAMMER                       Adolfo Di Mare    }

USES
  Dos, Crt, LngTools;


PROCEDURE WriteUSAGE;
{ RESULT
  Displays a help message. }
BEGIN { WriteUSAGE }
  WriteLn('LNGrep v', LngTools.LNG_version:1);
  WriteLn('USAGE: ');
  Write  ('LNGrep wildcard[.ext] [options] ');
  WriteLn('[LNG(lang)] [>file]');
  WriteLn;
  WriteLn(
  'Extracts from all the xBase[.prg] program files in');
  WriteLn(
  'the current directory all lines that contain text');
  WriteLn(
  'strings, and writes them to the standard output.');
END; { WriteUSAGE }

VAR
  SG : TString_Group;
  { SG is so big that it would blow up the run }
  { time stack if it were a local variable...  }

PROCEDURE Grep_File(                    { ADOLFO }
  {+} VAR outF : TEXT;      { output TEXTFILE    }
  {+} def_lng  : LNG_T;     { default annotation }
  {+} add_LNG  : BOOLEAN; { add LNG(*) if needed }
  {+} only_new : BOOLEAN; { output only new SG's }
  {+} VAR dir  : Dos.DirStr;      { Dos.Fsplit() }
  {+} VAR fname: Dos.PathStr   { file to process }
);
{ RESULT
  Reads the text file "fname", line by line,
  looking for xBase strings. When one is found,
  it writes the line once or twice on the
  standard output appending the && LNG()
  annotations if required.

  - outF must be open. It remains open upon
    return.                                      }

VAR
  inF : FILE_OF_CHAR_T;

  str : Tokenized_T;

  something_got_written : BOOLEAN;

  str_lng: STRING[LNG_size];

  i : WORD;

  line_no : WORD; { current line in "fname" }

  path: Dos.PathStr;

BEGIN { Grep_File }

  IF IS_bak(fname) THEN BEGIN
    { avoid overwriting backup files }
    EXIT;
  END;

  path := dir + fname;

  Assign(inF, path);
  Reset(inF);

  Write_File_Name(path, def_lng);

  { Initialize and pop empty the string group }
  SG.Init;

  { Main cicle: process one line per iteration }

  line_no := 0;
  something_got_written := FALSE;

  WHILE Get_Line(inF, str.str) DO BEGIN

    INC(line_no);
    Tokenize(str.str, str.prs, str.com, str.lng);

    IF (System.Pos(Quote,  str.str) = 0)
                       AND
       (System.Pos(QQuote, str.str) = 0)
    THEN BEGIN
      { the line does not have an Xbase string }
      SG.Write_Add(outF, def_lng, add_LNG,only_new,
                   something_got_written);
      { a new group begins: flush the old one }

      { just do nothing with the line }
    END

    ELSE BEGIN
      { the line has an Xbase string }

      IF str.lng = '' THEN BEGIN
        { clear the previous group }
        SG.Write_Add(outF, def_lng, add_LNG,only_new,
                     something_got_written);

        { there is no annotation: add it }
        IF NOT str.com THEN BEGIN
          SG.Insert(str, line_no);
        { SG.Write_Add(def_lng, add_LNG); }
        END;
      END

      ELSE IF str.lng = LNG_all THEN BEGIN
        { skip language independent strings }
        SG.Write_Add(outF, def_lng, add_LNG,only_new,
                     something_got_written);

      { this code is commented-out:
        IF NOT str.com THEN BEGIN
          SG.Insert(str, line_no);
          SG.Write_Add(def_lng);
          something_got_written := TRUE;
        END;
      }
        { LNG() lines don't get written out }
      END

      ELSE BEGIN
        { the line already has an LNG() annotation }

        IF (NOT str.com)      OR { not a comment?  }
           SG.Empty           OR { not in a group? }
           SG.IS_Member(str.lng) { already there?  }
        THEN BEGIN

          { a new group begins: flush the old one }
          SG.Write_Add(outF, def_lng, add_LNG,only_new,
                       something_got_written);

          { now decide whether to begin a new group }
          IF NOT str.com THEN BEGIN
            SG.Insert(str, line_no);
          END;

        END
        ELSE BEGIN
        { reverse condition:
          (str.com)      AND          it's a comment
          (NOT SG.Empty) AND          there is a group
          (NOT SG.IS_Member(str.lng)) not in the group
        }
          { add the line to the group }
          SG.Insert(str, line_no);
        END;

      END;

    END;   { the line has an Xbase string }

  END; { WHILE }

  { finish up the last one }
  SG.Write_Add(outF, def_lng, add_LNG,only_new,
               something_got_written);
  IF something_got_written THEN BEGIN
    Write_Dashes(outF, 0,0);
  END;

  IF NOT Eof(inF) THEN BEGIN
    Write  (outF, 'LNGrep: line too long in file ');
    WriteLn(outF,  fname, ' (', line_no, ')');
    WriteLn(outF, 'LNGrep: file scanning aborted!');
  END;

  SG.Done;
  Close(inF);
END; { Grep_File }


VAR
  def_lng  : LNG_T;
  add_LNG  : BOOLEAN;
  only_new : BOOLEAN;
  wildcard : STRING;


TYPE
  Turbo_SearchRec = RECORD       { from Online Help }
    fill: ARRAY[1..21] OF BYTE;
    attr: BYTE;
    time: LONGINT;
    size: LONGINT;
    name: STRING[12];
  END;


VAR
  srec  : Dos.SearchRec;
CONST
  {$IFNDEF VER55}   FName_S = SizeOf(srec.name);
  {$ELSE}           FName_S = 1+12;
  {$ENDIF !VER55}

TYPE
  Node_P = ^Node_T;
  Node_T = RECORD
    fname : STRING[FName_S];
    next  : Node_P;
  END;

VAR
  list,              { to chain up file names }
  node,
  last   : Node_P;

VAR
  dir : Dos.DirStr;
  path: Dos.PathStr; { to parse file names }
  name: Dos.NameStr;
  ext : Dos.ExtStr;

BEGIN { LNGrep }

  { get the command line parameters }
  IF NOT Command_Line(
           def_lng, wildcard, add_LNG, only_new)
  THEN BEGIN
    WriteUSAGE;
    EXIT;
  END;

  { redirect I/O }
  IF NOT ISaTTY(1) THEN BEGIN
    Assign(OUTPUT,'');
    ReWrite(OUTPUT);
  END;

  { parse the file name }
  Dos.Fsplit(wildcard, dir, name, ext);

  { "dir" contains the drive/directory }
  { "dir" is the path to each file     }

  { create a list with all the file names }
  Dos.FindFirst(
    wildcard,
    Dos.Archive + Dos.ReadOnly,
    srec);
  list := NIL;

  WHILE Dos.DosError = 0 DO BEGIN

    NEW(node);
    node^.fname := srec.name;

    IF list = NIL THEN BEGIN
      list := node;   { empty list }
    END
    ELSE BEGIN
      last^.next := node; { append }
    END;

    last := node;

    Dos.FindNext(srec);

  END;

  { finish up the list }
  IF list <> NIL THEN BEGIN
    last^.next := NIL;
  END;
{
  In LNGswap.pas I included an explanation of why
  I keep all the file names in a list, if you are
  interested.

  This really isn't necessary here, but I decided
  to include just in case.
}

  { process each file }
  WHILE list <> NIL DO BEGIN

    Grep_File(
      OUTPUT, def_lng, add_LNG, only_new,
      dir, list^.fname);

    node := list;
    list := list^.next;
    DISPOSE(node);
  END;

END. { LNGrep }
