{$A+ Align data on word boundary }
{$F- Don't force FAR calls       }
{$B- Short circuit boolean eval  }
{$V- Relax string checking       }

{$I- Enable System.IOResult error checks }
{$N+,E+ Allow for software 80x87 }
{$R- Range checking off }
{$S- Stack checking off }

{$L+ Enable gen. of local symbols }
{$D+ Enable debug information     }
{$O- No overlays                  }

{$M  4096,0, 655360} { Max-Stack Low-heap High-heap }

UNIT LNGtools; { (C) Copyright: Adolfo Di Mare, 1994 }

{ NOTICE
  Use this programs at your own risk. The author
  reserves all comercial rights. This program, its
  source code and documentation have been put in the
  public domain.                                     }

{ RESULT
  This unit contains several helper procedures for
  programs LNGrep, LNGstuff, and LNGswap.            }

{ PROGRAMMER                       Adolfo Di Mare    }

{$IFDEF VER40}
   ERROR: OOP support required!
{$ENDIF VER40}

{$IFDEF VER50}
   ERROR: OOP support required!
{$ENDIF VER50}

{[])===========([]}
{[]} INTERFACE {[]}
{[])===========([]}

USES
  Dos;

CONST
  Default_LNG_File = 'TEXTFILE';
  LNG_version = '1.2';
  LNG_default = 'LNG(us)';
  LNG_all     = 'LNG()';
  LNG_size    = 20;
  LineLength  = 210;
  Quote       = ''''; { ' }
  QQuote      = '"';  { " }
  Blank       = ' ';
  Dash        = '-';
  Comment     = '  && ';
  ERROR_line  = '<<******* ERROR *******>>';


TYPE
  FILE_OF_CHAR_T = FILE OF CHAR;
  LNG_T          = STRING[LNG_size];

FUNCTION  ISaTTY(         handle : WORD) : BOOLEAN;
FUNCTION  File_Exists(    VAR path : STRING): BOOLEAN;
FUNCTION  IS_Directory(   VAR path : STRING): BOOLEAN;
FUNCTION  IS_bak(         VAR path : STRING): BOOLEAN;
FUNCTION  UpperStr(       s: STRING): STRING;
FUNCTION  LowerStr(       s: STRING): STRING;
FUNCTION  RTrim(          s: STRING): STRING;
FUNCTION  Skip_Blanks(    VAR str: STRING): WORD;
FUNCTION  First_WORD(     VAR str: STRING; w: STRING)
                          : BOOLEAN;

PROCEDURE Str_Del_CHAR(   VAR str: STRING; ch: CHAR);

FUNCTION  Command_Line(   VAR lng : STRING;
                          VAR wildcard: STRING;
                          VAR add_LNG : BOOLEAN;
                          VAR only_new: BOOLEAN)
                          : BOOLEAN;

PROCEDURE Rename_to_BAK(  VAR path: Dos.PathStr;
                          VAR ext : Dos.ExtStr);

FUNCTION  Get_Line(       VAR f  : FILE_OF_CHAR_T;
                          VAR str: STRING) : BOOLEAN;

PROCEDURE Write_Dashes(   VAR F: TEXT; n,m : WORD);
FUNCTION  IS_Dashed(      VAR str: STRING;
                          VAR n,m: WORD  ) : BOOLEAN;
FUNCTION  IS_Blank(       VAR str: STRING) : BOOLEAN;

PROCEDURE Write_File_Name(VAR fname: STRING;
                          VAR def_lng:LNG_T);
FUNCTION  IS_File_Name(   VAR str  : STRING;
                          VAR fname: STRING): BOOLEAN;

FUNCTION  IS_Comment(     VAR str: STRING;
                          VAR pos: WORD): BOOLEAN;
FUNCTION  Has_Annotation( VAR str:STRING;VAR pos:WORD;
                          VAR lng: LNG_T): BOOLEAN;

PROCEDURE Tokenize(       VAR str,prs: STRING;
                          VAR com: BOOLEAN;
                          VAR lng: LNG_T);

TYPE
  Tokenized_P = ^Tokenized_T;
  Tokenized_T = RECORD
    com : BOOLEAN;              { Am I a comment? }
    str : STRING;           { the original string }
    prs : STRING;    { no comment + no annotation }
    lng : STRING[LNG_size];     { LNG(annotation) }
  {
    This record contains the original string, as read
    from the input file, and its parsed version.
    - If "str" is a comment, then "prs" will not
      have the leading comment characters in "str".
    - If "str" has a LNG(annotation), even if it is
      the empty LNG() annotation, "prs" will not
      have it.
    - "com" is TRUE when "str" is a comment.

    All these values can be set invoking procedure:
       Tokenize(str, prs, com, lng);
  }
  END;

CONST
  DIM_SG = 100; { This many Languages: 'you nuts??? }

{$UNDEF NO_PUBLIC}

{$IFDEF VER55}
  {$DEFINE NO_PUBLIC}
{$ENDIF VER55}

{$IFDEF VER60}
  {$DEFINE NO_PUBLIC}
{$ENDIF VER60}

TYPE
  TString_Group = OBJECT
    {[]========[]}                {$IFNDEF NO_PUBLIC}
    {|} PUBLIC {|}                  { COMMENT OUT }
    {[]========[]}                {$ENDIF  NO_PUBLIC}
    CONSTRUCTOR Init;
    PROCEDURE   Clear;
    DESTRUCTOR  Done;

    PROCEDURE Insert(     VAR s       : Tokenized_T;
                              line_no : WORD);

    PROCEDURE Write_One(  VAR F       : TEXT;
                          VAR lng : Lng_T);

    PROCEDURE Write_Add(  VAR F       : TEXT;
                          VAR lng     : Lng_T;
                              add_LNG : BOOLEAN;
                              only_new: BOOLEAN;
                          VAR wrote_some: BOOLEAN);
    PROCEDURE Write(      VAR F       : TEXT;
                              lng     : Lng_T);

    FUNCTION  IS_Member(      lng : LNG_T) : BOOLEAN;
    FUNCTION  IS_Similar( VAR s   : STRING): BOOLEAN;

    FUNCTION  Empty : BOOLEAN;
    FUNCTION  Count : WORD;

    {[]=========[]}                 {$IFNDEF VER55}
    {|} PRIVATE {|}                 { COMMENT OUT }
    {[]=========[]}                 {$ENDIF  VER55}
    line,          { where the group begins }
    cnt : WORD;         { # of used entries }
    str : ARRAY[0..DIM_SG-1] OF Tokenized_T;

  END;


{[]================[]}
{|} IMPLEMENTATION {|}
{[]================[]}


FUNCTION ISaTTY(                   { ADOLFO }
  {+} handle : WORD
) {>>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "handle" is the console,
  a serial port, a paralell port, or the
  keyboad.                                  }
VAR
  r : Dos.Registers;
BEGIN { ISaTTY }
  r.ax := $4400;
  r.bx := handle;
  MsDos(r);
  ISaTTY :=
    (0<>r.dx AND $8000) AND (0 = r.flags AND FCarry);
END; { ISaTTY }

FUNCTION File_Exists(          { ADOLFO }
  {+} VAR path : STRING
) {>>>>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when file "path" exists. }
VAR
  srec  : Dos.SearchRec;
BEGIN { File_Exists }
  Dos.FindFirst(path, Dos.AnyFile, srec);
  File_Exists := (Dos.DosError = 0);
END; { File_Exists }

FUNCTION IS_Directory(           { ADOLFO }
  {+} VAR path : STRING
) {>>>>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "path" is a directory.
  - It just checks whether "path\nul"
    exists.                               }
VAR
  srec : Dos.SearchRec;
BEGIN { IS_Directory }
  IF path[Length(path)]='\' THEN BEGIN
    Dos.FindFirst(path+'NUL',  Dos.AnyFile, srec);
  END
  ELSE BEGIN
    Dos.FindFirst(path+'\NUL', Dos.AnyFile, srec);
  END;
  IS_Directory := (Dos.DosError = 0);
END; { IS_Directory }

FUNCTION IS_bak(                 { ADOLFO }
  {+} VAR path : STRING
) {>>>>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "path" is a .bak file }
VAR
  dir : Dos.DirStr;
  name: Dos.NameStr;
  ext : Dos.ExtStr;
BEGIN { IS_bak }
  Dos.Fsplit(path, dir, name, ext);
  IS_bak := (ext = '.BAK')
END; { IS_bak }

FUNCTION UpperStr(          { ADOLFO }
  {+} s: STRING
) {>>>}: STRING;
{ RESULT
  Returns an uppercase copy of "s".  }
VAR
  i: INTEGER;
BEGIN { UpperStr }
  FOR i := 1 TO Length(s) DO BEGIN
     CASE s[i] OF
       '': s[i] := '';
       '': s[i] := '';
       '': s[i] := '';
       '': s[i] := '';
       '': s[i] := '';
     ELSE
       s[i] := System.UpCase(s[i]);
     END;
  END;
  UpperStr := s;
END; { UpperStr }

FUNCTION LowerStr(          { ADOLFO }
  {+} s: STRING
) {>>>}: STRING;
{ RESULT
  Returns a lower case copy of "s".  }
VAR
  i: INTEGER;
BEGIN { LowerStr }
  FOR i := 1 TO Length(s) DO BEGIN

     IF ('A' <= s[i]) AND (s[i] <= 'Z') THEN BEGIN
       s[i] := CHR( ORD('a') + (ORD(s[i])-ORD('A')) );
     END

     ELSE BEGIN
       CASE s[i] OF
         '': s[i] := '';
         '': s[i] := '';
         '': s[i] := '';
         '': s[i] := '';
         '': s[i] := '';
       END;
    END;
  END;

  LowerStr := s;
END; { LowerStr }

FUNCTION RTrim(          { ADOLFO }
  {+} s: STRING
) {>>>}: STRING;
{ RESULT
  Returns a copy of "s" where all
  trailing blanks are removed.    }
VAR
  n: WORD;
BEGIN { RTrim }
  n := WORD(s[0]);
  WHILE (n>0) DO BEGIN
    IF s[n] = Blank THEN BEGIN
      DEC(n);
    END
    ELSE BEGIN
      s[0] := CHAR(n);
      n := 0; {BREAK}
    END;
  END;

  RTrim := s;
END; { RTrim }

FUNCTION Skip_Blanks(       { ADOLFO }
  {+} VAR str: STRING
) {>>>>>>>>>}: WORD;
{ RESULT
  Returns the position of the first
  non blank character in "str".
  - If "str" is all blanks, it
    returns Length(str)+1.
  - If "str" is null, it returns 1.  }
VAR
  i,
  mark : WORD;
BEGIN { Skip_Blanks }

  i := 1;
  mark := Length(str);
  WHILE i <= mark DO BEGIN
    IF str[i] <> Blank THEN BEGIN
      mark := 0;
    { BREAK; }
    END
    ELSE BEGIN
      INC(i);
    END;
  END;

  Skip_Blanks := i;
END; { Skip_Blanks }

FUNCTION First_WORD(                { ADOLFO }
  {+} VAR str : STRING;
  {+}     w   : STRING
) {>>>>>>>>}: BOOLEAN;
{ RESULT
  Returns TRUE is the first word in "str" is
  "aword".
  - leading blanks in both "str" and "w" are
    stripped off before comparison.
  - The comparison is done ignoring case.    }
VAR
  i, maxI : WORD;
  j, maxJ : WORD;

BEGIN { First_WORD }

  i := Skip_Blanks(str);
  j := Skip_Blanks(w);

  maxI := Length(str);
  maxJ := Length(w);

  WHILE (i <= maxI) AND (j <= maxJ) DO BEGIN
    IF UpCase(str[i]) <> UpCase(w[j]) THEN BEGIN
      maxI := 0;
    { BREAK; }
    END;
    INC(i);
    INC(j);
  END;

  First_WORD := (j > maxJ);

END; { First_WORD }

PROCEDURE Str_Del_CHAR(      { ADOLFO }
  {?} VAR str : STRING;
  {+}     ch  : CHAR
);
{ RESULT
  Removes from "str" every occurence  }
{ EXAMPLE
  :before:
  <c is a language for your computers>
  :after:
  < is a language for your omputers>  }
VAR
  sz,
  i,j : WORD;
BEGIN { Str_Del_CHAR }
  i := 1;
  j := 0;
  sz := Length(str);

  WHILE i <= sz DO BEGIN
    IF ch = str[i] THEN BEGIN
      { skip }
    END
    ELSE BEGIN
      { copy }
      INC(j);
      str[j] := str[i];
    END;
    INC(i);
  END;

  str[0] := CHR(j);
END; { Str_Del_CHAR }

FUNCTION Command_Line(              { ADOLFO }
  {-} VAR lng     : STRING;
  {-} VAR wildcard: STRING;
  {-} VAR add_LNG : BOOLEAN;
  {-} VAR only_new: BOOLEAN
) {>>>>>>>>>>>>>>}: BOOLEAN;
{ RESULT
  Parses the command line to returns the
  language and wildcard parameters.
  - On error, writes a message on the
    console return FALSE.                    }
VAR
  i,n: WORD;
  str: STRING;
BEGIN { Command_Line }

  n := System.ParamCount;
  add_LNG  := FALSE;
  only_new  := FALSE;
  wildcard := '';

  IF NOT ( (1 <= n) AND (n <= 3) ) THEN BEGIN
    Command_Line := FALSE;
    EXIT;
  END;

  { Cicle to read the parameters in the command line }
  FOR i := 1 TO n DO BEGIN

    { get next parameter }
    str := UpperStr(System. ParamStr(i));

    IF System.Copy(str,1,4) = 'LNG(' THEN BEGIN

      str :=
        System.Copy(str,5, System.Pos(')',str)-1-4);

      IF str = '' THEN BEGIN
        Write('LNG(): ERROR in language parameter');
        EXIT;
      END
      ELSE BEGIN
        lng := 'LNG(' + LowerStr(str) + ')';
      END;
      IF lng = LNG_all THEN BEGIN
        Write(
          '"', LNG_all,
          '"is an invalid language antotation');
        Command_Line := FALSE;
        EXIT;
      END;
    END

    ELSE IF
      (str = '-H')  OR  (str = '/H')  OR
      (str = '-?')  OR  (str = '/?')
    THEN BEGIN
      Command_Line := FALSE;
      EXIT;
    END

    ELSE IF
      (str = '/ADD') OR
      (str = '+ADD') OR
      (str = '-ADD')
    THEN BEGIN
      add_LNG := TRUE;
    END

    ELSE IF
      (str = '/NEW') OR
      (str = '+NEW') OR
      (str = '-NEW')
    THEN BEGIN
      only_new := TRUE;
    END

    ELSE BEGIN
      wildcard := str;
    END;

  END; { FOR }

  IF lng = '' THEN BEGIN
    lng := LNG_default;
  END;

  IF wildcard = '' THEN BEGIN
    { no wildcard file spec present }
  END

  ELSE IF
    (System. Pos('\', wildcard) <> 0) OR
    (System. Pos('/', wildcard) <> 0) OR
    (System. Pos(':', wildcard) <> 0)
  THEN BEGIN
    { a wildcard file spec present }
  END;

  Command_Line := TRUE;
END; { Command_Line }

PROCEDURE Rename_to_BAK(               { ADOLFO }
  {+} VAR path: Dos.PathStr;
  {+} VAR ext : Dos.ExtStr
);
{ RESULT
  Renames "path.ext" to "path.bak".
  - If there is already a .bak file for "name",
    then it gets deleted.
  - "name" can not be a path name.              }

VAR
  f : FILE;
  bak: STRING[SizeOf(Dos.PathStr)];

BEGIN { Rename_to_BAK }

  { erase .bak }
  bak := path + '.bak';
  IF File_Exists(bak) THEN BEGIN
    Assign(f, bak);
    Erase(f);
  END;

  { rename to .bak }
  Assign(f, path+ext);
  Rename(f, bak);
END; { Rename_to_BAK }


FUNCTION Get_Line(              { ADOLFO }
  {+} VAR f   : FILE_OF_CHAR_T;
  {-} VAR str : STRING
) {>>>>>>>>}  : BOOLEAN;
{ RESULT
  Gets a line from file "f".
  - When the line to get is too long, it
    returns FALSE, and returns as much
    of the line as it can. In this case
    all the other characters in the line
    get skiped.
  - Get_Line() reads "f" char by char,
    so it helps a lot if "f" is buffered.
  - On Eof(f), Get_Line returns FALSE.   }
{ REMARKS
  The only reason for not using ReadLn()
  instead of Get_Line() is to protect a
  [naive] user from throwing to the
  LNGrep program lines with more than
  255 characters. 
  - A work around is to reprogram the 
    whole thing in C++, or to wait 2 
    centuries until computers with no
    limits are in wide use.              }
VAR
  i : WORD;
  ch: CHAR;

  done: BOOLEAN;

BEGIN { Get_Line }

  IF Eof(f) THEN BEGIN
    Get_Line := FALSE;
    EXIT;
  END;

  i    := 0;
  done := FALSE;
  WHILE NOT done DO BEGIN

    INC(i);
    IF Eof(f) THEN BEGIN
      done := TRUE;
    END
    ELSE BEGIN
      Read(f, ch);

      IF ch = CHR(13) THEN BEGIN   { Ctrl-M  }
        Read(f, ch);               { skip LF }
        done := TRUE;
      END
      ELSE IF i>LineLength THEN BEGIN
        REPEAT
          Read(f, ch);      { skip all chars }
        UNTIL (ch = CHR(13)) OR Eof(f);
        IF NOT Eof(f) THEN BEGIN
          Read(f, ch);        { skip LF }
        END;
        done := TRUE;
      END
      ELSE BEGIN
        str[i] := ch;
      END;
    END;

  END;

  str[0] := CHR(i-1);
  Get_Line := (i<=LineLength);
END; { Get_Line }


PROCEDURE Write_Dashes(      { ADOLFO }
  {?} VAR F : TEXT;
  {+}     n,
  {+}     m : WORD
);
{ RESULT
  Prints on standard output a lot of
  pretty dashes.
  - Used to separate groups of lines. }
VAR
  i,
  ndig,
  mdig : WORD;
BEGIN { Write_Dashes }
  Write(F, '[]');

  { ndig the number of digits in line_no }
  i    := n;
  ndig := 0;
  REPEAT
    i := i DIV 10;
    INC(ndig);
  UNTIL i = 0;

  i    := m;
  mdig := 0;
  REPEAT
    i := i DIV 10;
    INC(mdig);
  UNTIL i = 0;

  FOR i := 1 TO 78-2*Length('[]')-(ndig+mdig+1)
  DO BEGIN
    Write(F, Dash);
  END;

  WriteLn(F, '[', n:ndig, ':', m:mdig, ']');
END; { Write_Dashes }

FUNCTION IS_Dashed(             { ADOLFO }
  {+} VAR str : STRING;
  {-} VAR n,               { line number }
  {-}     m   : WORD       { # in group  }
) {>>>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "str" is a line that
  was printed with Write_Dashes.
  - TRUE when str has the form:
    []--------------------------[nnn:mm]
    where, "nnn" and "mm" are 0 or more
    digits.
  - The function is quite liberal,
    because it does no requiere "str" to
    have an exact number of dashes.
  - On exit, if the line is dashed, then
    "n" and "m" will contain the numeric
    values that appear at the end of the
    line. Otherwise, both variables will
    remain unchanged.                    }
VAR
  i,
  mark,
  last : WORD;

BEGIN { IS_Dashed }
  { find leading  '[]' }
  i := Skip_Blanks(str);
  IF (str[i] <> '[') OR (str[i+1] <> ']') THEN BEGIN
    IS_Dashed := FALSE;
    EXIT;
  END;

  { skip dashes }
  last := Length(str);
  mark := last;
  i    := i+2;  { skip '[]' }

  WHILE i<mark DO BEGIN
    IF str[i] <> Dash THEN BEGIN
      mark := i;
    END
    ELSE BEGIN
      INC(i);
    END;
  END;

  IF str[i] <> '[' THEN BEGIN
    IS_Dashed := FALSE;
    EXIT;
  END;

  INC(i);  { skip '[' }

  { gobble up digits }
  last := 0;
  WHILE ('0' <= str[i]) AND (str[i] <= '9') DO BEGIN
    { build up numeric quantity }
    last := last * 10 + Ord(str[i])-Ord('0');
    INC(i);
  END;

  IF str[i] <> ':' THEN BEGIN
    IS_Dashed := FALSE;
    EXIT;
  END;

  INC(i);  { skip ':' }

  { gobble up digits }
  mark := 0;
  WHILE ('0' <= str[i]) AND (str[i] <= '9') DO BEGIN
    { build up numeric quantity }
    mark := mark * 10 + Ord(str[i])-Ord('0');
    INC(i);
  END;

  IF str[i] = ']' THEN BEGIN
    n := last; 
    m := mark;
    IS_Dashed := TRUE;
  END
  ELSE BEGIN
    IS_Dashed := FALSE;
  END;

END; { IS_Dashed }

FUNCTION IS_Blank(                  { ADOLFO }
  {+} VAR str : STRING
) {>>>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "str" is blank or empty. }
VAR
  i,
  mark : WORD;

BEGIN { IS_Blank }

  i := 1;
  mark := Length(str);

  WHILE i <= mark DO BEGIN
    IF str[i] <> Blank THEN BEGIN
      i := mark+1;
    END;
    INC(i);
  END;

  IS_Blank := (i <> mark+2);
END; { IS_Blank }

PROCEDURE Write_File_Name(      { ADOLFO }
  {+} VAR fname  : STRING;
  {+} VAR def_lng: LNG_T
);
{ RESULT
  Writes in the standard output "fname",
  in a format that IS_File_Name() can
  recognize.
  - The format is:
    [>====> File [name.exe]              }
BEGIN { Write_File_Name }
  WriteLn(
    '*==>  File  [', fname, ']',
    Comment, def_lng);
END; { Write_File_Name }

FUNCTION IS_File_Name(          { ADOLFO }
 {+} VAR str  : STRING;
 {-} VAR fname: STRING
){>>>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "str" is the first
  line that contains the file name where
  for a group of xBase strings.
  - TRUE when str has the form:
    [>====> File [d:path\name.ext]
  - When "str" is a file name string, the
    file name is returned in "fname".
    Otherwise, "fname" is null.
  - White space within "str" is ignored. }
VAR
  i,j,
  mark,
  last : WORD;
  wFILE: STRING[4];

BEGIN { IS_File_Name }
  fname := '';

  { find leading  '*' }
  i := Skip_Blanks(str);
  IF (str[i] <> '*') THEN BEGIN
    IS_File_Name := FALSE;
    EXIT;
  END;

  last := Length(str);

  { skip '====' }
  mark := last;
  i    := i+1;   { skip '>' }
  WHILE i<mark DO BEGIN
    IF str[i] <> '=' THEN BEGIN
      mark := i;
    END
    ELSE BEGIN
      INC(i);
    END;
  END;

  IF (str[i] <> '>') THEN BEGIN
    IS_File_Name := FALSE;
    EXIT;
  END;

  { skip Blanks }
  mark := last;
  i    := i+1;  { skip '>' }
  WHILE i<mark DO BEGIN
    IF str[i] <> Blank THEN BEGIN
      mark := i;
    END
    ELSE BEGIN
      INC(i);
    END;
  END;

  { test for 'File' }
  wFILE := 'FILE';
  j     := 1;
  WHILE j<5 DO BEGIN
    IF UpCase(str[i]) = wFILE[j] THEN BEGIN
      INC(i);
      INC(j);
    END
    ELSE BEGIN
      j := 5+1;
    END;
  END;

  IF j > 5 THEN BEGIN
    IS_File_Name := FALSE;
    EXIT;
  END;

  { skip Blanks }
  mark := last;
  WHILE i<mark DO BEGIN
    IF str[i] <> Blank THEN BEGIN
      mark := i;
    END
    ELSE BEGIN
      INC(i);
    END;
  END;

  IF str[i] <> '[' THEN BEGIN
    IS_File_Name := FALSE;
    EXIT;
  END;

  { gobble up the file name until done }
  mark := last;
  i    := i+1;  { skip ']' }
  WHILE i<=mark DO BEGIN
    IF str[i] <> ']' THEN BEGIN
      fname := fname + str[i];
    END
    ELSE BEGIN
      mark := 0;
    END;
    INC(i);
  END;

  IS_File_Name := TRUE;

END; { IS_File_Name }

FUNCTION IS_Comment(                 { ADOLFO }
  {+} VAR str: STRING;
  {-} VAR pos: WORD
) {>>>>>>>>>}: BOOLEAN;
{ RESULT
  Returns TRUE when "str" is an Xbase comment.
  - When "str" is a comment, "pos" contains
    the position where the comment characters
    are.
  - Otherwise, the value in "pos" remains
    unchanged.                                }
VAR
  i,
  mark : WORD;
  done : BOOLEAN;
BEGIN { IS_Comment }
  { skip leading blanks }
  i    := 0;
  mark := Length(str);
  done := FALSE;
  WHILE NOT done DO BEGIN
    INC(i);
    IF str[i] = Blank THEN BEGIN
      done := (i>mark);
    END
    ELSE BEGIN
      done := TRUE;
    END;
  END;

  IS_Comment := FALSE;
  IF str[i] = '*' THEN BEGIN  { leading '*' }
    pos := i;
    IS_Comment := TRUE;
  END
  ELSE BEGIN
    IF i<mark THEN BEGIN      { leading '&&' }
      IF (str[i]='&') AND (str[i+1]='&') THEN BEGIN
        pos := i;
        IS_Comment := TRUE;
      END;
    END;
  END;
END; { IS_Comment }

FUNCTION Has_Annotation(         { ADOLFO }
  {+} VAR str: STRING;           { SELF   }
  {-} VAR pos: WORD;             { where  }
  {-} VAR lng: LNG_T         { annotation }
) {>>>>>>>>>}: BOOLEAN;
{ RESULT
  Returns TRUE when "str" has a language
  annotation at the end.
  - When "str" has the annotation, "pos"
    contains the position where the
    annotation begins, and lng contains
    the annotation.
  - Otherwise, (lng='') AND (pos=0).      }
VAR
  i,
  mark : WORD;
BEGIN { Has_Annotation }
  { find the last && }
  pos := 0;
  i   := Length(str);

  WHILE i > 0 DO BEGIN

    IF str[i] = '&' THEN BEGIN
      IF i > 1 THEN BEGIN
        IF str[i-1] = '&' THEN BEGIN
          pos := i-1;
          i   := 1;      {BREAK;}
        END;
      END
    END;

    DEC(i);
  END;

  IF pos = 0 THEN BEGIN
    lng := ''; { sorry: no annotation }
    Has_Annotation := FALSE;
  END
  ELSE BEGIN
{   mark := pos; }

    lng := System.Copy(str, pos, Length(str)-pos+1);
    lng := UpperStr(lng);

    mark := System.Pos('LNG(', lng);
    i    := System.Pos(')', lng);

    IF (mark = 0) OR (mark >= i) THEN BEGIN
      lng := '';
      pos := 0;
      Has_Annotation := FALSE;
    END
    ELSE BEGIN
      lng :=
        System.Copy(lng, mark+4, i-(mark+4));

      lng := 'LNG(' + LowerStr(lng) + ')';

      Has_Annotation := TRUE;
    END;
  END;
END; { Has_Annotation }

PROCEDURE Tokenize(                   { ADOLFO }
  {+} VAR str,              { original string  }
  {-}     prs : STRING;     { parsed string    }
  {-} VAR com : BOOLEAN;    { am I a comment?  }
  {-} VAR lng : LNG_T       { LNG() annotation }
);
{ RESULT
  Parses "str" and extracts its LNG() annotation
  when one is present.
  - "str" remains unchanged.
  - "prs" is a copy of "str" with the LNG()
    annotation and the comment removed.
  - When "str" is a commented out Xbase program
    line, "com" is set to TRUE, and the comment
    character[s] is[are] set to Blank.
  - If "str" has an LNG() annotation, it gets
    stripped out of it.
  - Trailing blanks are not removed from
    "prs".                                     }
VAR
  i,
  mark,
  truncate : WORD;

  temp: LNG_T;
  done: BOOLEAN;
BEGIN { Tokenize }
  { copy the sucker }
  prs := str;

  { find out if the line is a comment line }
  com := IS_comment(prs, i);

  { get rid of the comment characters }
  IF com THEN BEGIN
    IF prs[i] <> '*' THEN BEGIN
      prs[i+1] := Blank;
    END;
    prs[i] := Blank;
  END;

  { figure out where to truncate prs }
  IF Has_Annotation(prs, truncate, lng) THEN BEGIN
    { this code will strip out trailing blanks: }
    {
      REPEAT
        DEC(truncate);
      UNTIL prs[truncate] <> Blank;
    }

    { jump before && }
    DEC(truncate);

    { strip out blanks leading the annotation }
    temp := Comment;
    i    := Skip_Blanks(temp)-1;
    WHILE i > 0 DO BEGIN
      IF prs[truncate] = Blank THEN BEGIN
        DEC(truncate);
        DEC(i);
      END
      ELSE BEGIN
        i := 0;
      END;
    END;

    { strip out annotation }
    prs[0] := CHR(truncate);
  END;

END; { Tokenize }

CONSTRUCTOR TString_Group.Init;  { ADOLFO }
{ RESULT
  Initializes "SELF".                     }
BEGIN { Init }
  SELF.cnt := 0;
END; { Init }

PROCEDURE TString_Group.Clear;   { ADOLFO }
{ RESULT
  Removes every string from "SELF".       }
BEGIN { Clear }
  SELF.cnt := 0;
END; { Clear }

DESTRUCTOR TString_Group.Done;   { ADOLFO }
{ RESULT
  Destroys "SELF".                        }
BEGIN { Done }
{
  Its good discipline to always destroy
  constructed objets.
}
END; { Done }

PROCEDURE TString_Group.Insert (  { ADOLFO }
  {+} VAR s : Tokenized_T;     { insert me }
  {+}     line_no : WORD
);
{ RESULT
  Adds "s" to the token array "SELF".
  - No check for uniqueness is made, so
    it is posible for varios copies of "s"
    to be inserted in "SELF".
  - When "SELF" is empty, then "line_no"
    is used as the line number for the
    group.                                 }

BEGIN { Insert }
  IF SELF.cnt >= SizeOf(SELF.str)/SizeOf(Tokenized_T)
  THEN BEGIN
    WriteLn('*** OUT OF MEMORY ***');
    RunError( 201 {Range check error});
    EXIT;
  END;

  IF SELF.cnt = 0 THEN BEGIN { Empty? }
    SELF.line := line_no;
  END;

  { add str }
  SELF.str[SELF.cnt] := s;
  INC(SELF.cnt);
END; { Insert }

FUNCTION TString_Group.IS_Member( { ADOLFO }
  {+} lng : LNG_T
) {>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "SELF" already has a
  string with a LNG() annotation equal to
  "lng".                                   }
VAR
  i,
  mark : WORD;
BEGIN { IS_Member }
  i    := 0;
  mark := SELF.cnt;
  WHILE i < mark DO BEGIN
    IF SELF.str[i].lng = lng THEN BEGIN
      mark := i;
    { BREAK; }
    END;
    INC(i);
  END;

  IS_Member := (mark < SELF.cnt);
END; { IS_Member }

FUNCTION TString_Group.IS_Similar( { ADOLFO }
  {+} VAR s : STRING
) {>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "SELF" has a string
  similar to "s".
  - Two strings are considered similar if
    after removing from both of them the
    LNG() annotation, the comment character
    and all their blanks, they compare
    equal.                                  }
VAR
  i,
  mark   : WORD;

  temp,
  prs : STRING;
  com : BOOLEAN;
  lng : LNG_T;
BEGIN { IS_Similar }
  Tokenize(s, prs, com, lng);

  Str_Del_CHAR(prs, Blank);

  i    := 0;
  mark := SELF.cnt;
  WHILE i < mark DO BEGIN

    temp := SELF.str[i].prs;
    Str_Del_CHAR(temp, Blank);

    IF temp = prs THEN BEGIN
      mark := i;
    { BREAK; }
    END;

    INC(i);
  END;

  IS_Similar := (mark < SELF.cnt);
END; { IS_Similar }

FUNCTION TString_Group.Empty  { ADOLFO }
  {>>>>>>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "SELF" does not
  contain any strings.                 }
BEGIN { Empty }
  Empty := (SELF.cnt = 0);
END; { Empty }

FUNCTION TString_Group.Count  { ADOLFO }
  {>>>>>>>>>} : WORD;
{ RESULT
  Returns the number of line that group
  "SELF" contains.                     }
BEGIN { Count }
  Count := SELF.cnt;
END; { Count }


PROCEDURE TString_Group.Write_One(        { ADOLFO }
  {?} VAR F   : TEXT;
  {?} VAR lng : Lng_T
);
{ RESULT
  Writes out the lines in group "SELF" that has
  an annotation equal to "lng".
  - IF "SELF" is empty nothing gets written.
  - The annotation at the end of the line gets
    eliminated.
  - When the command line annotation "lng" then
    the first string in the group is written
    out.
  - This method is the little brother to
    SG.Write_Add(), because it does almost the
    same, but writes out only one line from
    the string group.
  - If the line group contains only the
    language independent annotation, then it is
    written out.
  - SELF becomes emtpy after invoking this
    method.                                        }

VAR
  i,
  mark : WORD;
BEGIN { Write_One }
  { chicken out asap }
  IF SELF.cnt = 0 THEN BEGIN
    EXIT;
  END;

  { find where the LNG() command line annotation is }
  i    := 0;
  mark := SELF.cnt;
  WHILE i < mark DO BEGIN
    IF SELF.str[i].lng = lng THEN BEGIN
      mark := i;
      { mark points to the lng annotation in SELF }
    { BREAK; }
    END;
    INC(i);
  END;

  { mark < SELF.cnt <==> "lng" appears in SELF   }
  { ie, the command line annotation is at "mark" }


  { Write the annotation for "lng" }
  IF (mark < SELF.cnt) THEN BEGIN
    System.Write(F, SELF.str[mark].prs);
  END
  ELSE BEGIN
    { there was no lng annotation: }
    { copy the first line          }
    System.Write(F, SELF.str[0].prs);
  END;

  System.WriteLn(F);

  { flush }
  SELF.cnt := 0;

END; { Write_One }

PROCEDURE TString_Group.Write_Add(    { ADOLFO }
  {?} VAR F       : TEXT;
  {?} VAR lng     : Lng_T;
  {+}     add_LNG : BOOLEAN;
  {+}     only_new: BOOLEAN;
  {?} VAR wrote_some: BOOLEAN
);
{ RESULT
  Writes out the lines in group "SELF".
  - IF "SELF" is empty nothing gets written.
  - All the lines but the first one are
    commented out.
  - Writes a dashed separator line before
    every line in the group.
  - The first string written out is the one
    which has the command language annotation
    equal to "lng".
  - When the command line annotation "lng" does
    not appear withing "SELF", it is included.
  - If "add_LNG" is TRUE, then an emtpy
    annotated string equal to the empty
    annotation is written out. In this case,
    the string that has the "lng" is used as
    the base to produce the annotated empty
    string.
  - If "only_new" is TRUE, then a string group
    is written out only if it does not have
    already an antotion equal to "lng", or if
    it doesn't have an annotation at all.
  - If the line group contains only the
    language independent annotation, then no
    line is added when "add_LNG" is TRUE.
  - When something gets written out, then
    variable "wrote_some" is changed to
    TRUE. Otherwise it is not changed.
  - SELF is emtpy after invoking this method.  }

VAR
  i,
  mark : WORD;
BEGIN { Write_Add }
  IF SELF.cnt = 0 THEN BEGIN
    EXIT;
  END;

  { find where the LNG() command line annotation is }
  i    := 0;
  mark := SELF.cnt;
  WHILE i < mark DO BEGIN
    IF SELF.str[i].lng = lng THEN BEGIN
      mark := i;
      { mark points to the lng annotation in SELF }
    { BREAK; }
    END;
    INC(i);
  END;

  { mark < SELF.cnt <==> "lng" appears in SELF   }
  { ie, the command line annotation is at "mark" }

  { decide whether not to write out the string group }
  IF              only_new
                    AND
     (
      ((SELF.cnt=1) AND (SELF.str[0].lng=LNG_all))
            OR
      ((SELF.cnt>1) AND (mark < SELF.cnt))
     )
  THEN BEGIN
    SELF.cnt := 0;    { flush }
    EXIT;
  END;

  wrote_some := TRUE;

  { Writes the dashed separator line }
  Write_Dashes(F, SELF.line, SELF.cnt);

  { add an empty annotation if needed }
  IF              add_LNG
                    AND
     (
      ((SELF.cnt=1) AND (SELF.str[0].lng<>LNG_all))
                       OR
                (SELF.cnt <> 1)
     )
  THEN BEGIN

    IF mark < SELF.cnt THEN BEGIN
      { copy the "lng" annotated line }
      System.Write(F, SELF.str[mark].prs);
    END
    ELSE BEGIN
      { copy the first line }
      System.Write(F, SELF.str[0].prs);
    END;

    System.Write(F, Comment, LNG_all);
    System.WriteLn(F);

  END;

  { Write the annotation for "lng" }
  IF (mark < SELF.cnt) THEN BEGIN
    System.Write(F, SELF.str[mark].prs);
  END
  ELSE BEGIN
    { there was no lng annotation: }
    { copy the first line         }
    System.Write(F, SELF.str[0].prs);
  END;

  System.Write(F, Comment, lng);
  System.WriteLn(F);
{
  When (SELF.str[i].lng = '') the string gets
  written out with the "lng" annotation.
  - In this case, it must not be written out
    again in the next loop.
}

  { write the other lines in the group }
  i := 0;
  WHILE i < SELF.cnt DO BEGIN
    IF (SELF.str[i].lng <> '')
                AND
            (mark <> i)
    THEN BEGIN

      System.Write(F, SELF.str[i].prs);
      System.Write(F, Comment);
      System.Write(F, SELF.str[i].lng);
      System.WriteLn(F);

    END;
    INC(i);
  END;

  { flush }
  SELF.cnt := 0;

END; { Write_Add }

PROCEDURE TString_Group.Write(        { ADOLFO }
  {?} VAR F   : TEXT;
  {+}     lng : LNG_T
);
{ RESULT
  Writes into "F" the lines in group "SELF".
  - When "SELF" is empty nothing gets written.
  - The first string from "SELF" written out is
    the one which has the command language
    annotation equal to "lng".
  - All the lines in "SELF" but the first one
    are commented out. Note that SG.Write_Add()
    never comments out the lines, because
    it's output goes to TEXTFILE, whereas the
    output of SG.Write() goes into a source
    [.prg] file.
  - Lines with the language independent
    annotation are not written out, unless the
    line group contains exactly one line.
  - If the command line annotation "lng" does
    not appear withing "SELF", it is not
    included.
  - SELF is not emtpy after invoking this
    method.                                    }

VAR
  i,
  mark : WORD;

  add_blank   : BOOLEAN;
  comment_out : CHAR;
BEGIN { Write }
  IF SELF.cnt = 0 THEN BEGIN
    EXIT;
  END;
  { Assert: SELF.cnt >= 1 }

  { force allignment for the string in the group }
  add_blank :=
    (SELF.str[0].str[1] <> Blank)
            AND
    (
      (SELF.cnt > 1)   OR
      ((SELF.cnt = 1) AND (SELF.str[0].lng <> lng))
    );

  { find where the LNG() command line annotation is }
  i    := 0;
  mark := SELF.cnt;
  WHILE i < mark DO BEGIN
    IF SELF.str[i].lng = lng THEN BEGIN
      mark := i;
      { mark points to the lng annotation in SELF }
    { BREAK; }
    END;
    INC(i);
  END;

  { mark < SELF.cnt <==> "lng" appears in SELF   }
  { ie, the command line annotation is at "mark" }

  IF (SELF.str[0].lng = LNG_all)
                AND
           (SELF.cnt = 1)
  THEN BEGIN
    { fake the "found" condition }
    mark := 0;
  END;

  IF (NOT SELF.IS_Member(lng))
               AND
        (mark >= SELF.cnt)
  THEN BEGIN
    { there is no lng annotation: make one up! }
    mark := SELF.cnt;
    INC(SELF.cnt);
    SELF.str[mark] := SELF.str[0];
    SELF.str[mark].lng := lng;
  END;

  { Write the annotation for "lng" }
  IF add_blank THEN BEGIN
    System.Write(F, Blank);
  END;

  System.Write(F, SELF.str[mark].prs);
  System.Write(F, Comment, SELF.str[mark].lng);
  System.WriteLn(F);

  { write the other lines in the group }
  i := 0;
  comment_out := '*';
  WHILE i < SELF.cnt DO BEGIN

    IF (SELF.str[i].lng <> LNG_all)
                   AND
               (mark <> i)
    THEN BEGIN

      { comment everybody out }
      IF add_blank THEN BEGIN
        System.Write(F, comment_out);
      END
      ELSE BEGIN
        SELF.str[i].prs[1] := comment_out;
      END;

      System.Write(F, SELF.str[i].prs);
      System.Write(F, Comment);
      System.Write(F, SELF.str[i].lng);
      System.WriteLn(F);
    END;

    INC(i);
  END;

  { flush }
  SELF.cnt := 0;

END; { Write }


BEGIN { LNGtools }
END.  { LNGtools }


FUNCTION IS_letter(         { ADOLFO }
  {+} ch : CHAR
) {>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "ch" is a letter }
BEGIN { IS_letter }
  ch := UpCase(ch);
  IS_letter := ('A' <= ch) AND (ch <= 'Z');
END; { IS_letter }

FUNCTION IS_digit(         { ADOLFO }
  {+} ch : CHAR
) {>>>>} : BOOLEAN;
{ RESULT
  Returns TRUE when "ch" is a digit }
BEGIN { IS_digit }
  IS_digit := ('0' <= ch) AND (ch <= '9');
END; { IS_digit }

